def set_deploy_parser(parser=None):
    if not parser:
        from .base import set_base_parser

        parser = set_base_parser()

    parser.add_argument(
        'path',
        type=str,
        help='The local path to a Jina flow project.',
    )
    parser.add_argument(
        '--name',
        type=str,
        help='If set, the URL of the flow will be prepended with this name',
    )
    parser.add_argument(
        '--workspace',
        type=str,
        help='The string ID of a workspace. If set, re-use the given workspace instead of creating a new one.',
    )
    return parser



import asyncio
import logging
import os
import shutil
import sys
import tempfile
import warnings
from contextlib import contextmanager
from pathlib import Path

__windows__ = sys.platform == 'win32'


def is_latest_version(package):
    def _parse_latest_release_version(resp):
        # credit: https://stackoverflow.com/a/34366589
        import json
        from packaging.version import parse

        latest_release_ver = parse('0')
        j = json.load(resp)
        releases = j.get('releases', [])
        for release in releases:
            latest_ver = parse(release)
            if not latest_ver.is_prerelease:
                latest_release_ver = max(latest_release_ver, latest_ver)
        return latest_release_ver

    try:
        import warnings
        from urllib.request import Request, urlopen
        import pkg_resources
        from packaging.version import Version
        from rich import print
        from rich.panel import Panel

        cur_ver = Version(pkg_resources.get_distribution(package).version)

        req = Request(
            f'https://pypi.python.org/pypi/{package}/json',
            headers={'User-Agent': 'Mozilla/5.0'},
        )
        with urlopen(
            req, timeout=5
        ) as resp:  # 'with' is important to close the resource after use
            latest_release_ver = _parse_latest_release_version(resp)
            if cur_ver < latest_release_ver:
                print(
                    Panel(
                        f'You are using [b]{cur_ver}[/b], but [green][b]{latest_release_ver}[/b][/green] is available. '
                        f'You may upgrade via [b]pip install -U {package}[/b]',
                        title=':new: New version available',
                        width=50,
                    )
                )
    except Exception as ex:
        # no network, too slow, PyPi is down
        get_logger().debug(ex)


def get_logger():
    from rich.logging import RichHandler

    logging.basicConfig(
        level=os.environ.get('JCLOUD_LOGLEVEL'),
        format='%(message)s',
        datefmt='[%X]',
        handlers=[RichHandler(rich_tracebacks=True)],
    )

    return logging.getLogger('jcloud')


def get_or_reuse_loop():
    """
    Get a new eventloop or reuse the current opened eventloop.

    :return: A new eventloop or reuse the current opened eventloop.
    """
    try:
        loop = asyncio.get_running_loop()
        if loop.is_closed():
            raise RuntimeError
    except RuntimeError:
        _update_policy()
        # no running event loop
        # create a new loop
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
    return loop


def _update_policy():
    if __windows__:
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

    elif 'JINA_DISABLE_UVLOOP' in os.environ:
        return
    else:
        try:
            import uvloop

            asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
        except ModuleNotFoundError:
            warnings.warn(
                'Install `uvloop` via `pip install uvloop` for better performance.'
            )


def get_pbar(description, disable=False, total=4):
    from rich.progress import (
        BarColumn,
        MofNCompleteColumn,
        Progress,
        SpinnerColumn,
        TextColumn,
        TimeElapsedColumn,
    )

    pbar = Progress(
        SpinnerColumn(),
        TextColumn('[bold]{task.fields[title]}'),
        '•',
        TextColumn('[dim]{task.description}...'),
        BarColumn(bar_width=20),
        MofNCompleteColumn(),
        '•',
        TimeElapsedColumn(),
        transient=True,
        disable=disable,
    )

    pb_task = pbar.add_task(description, total=total, start=False, title='')
    return pbar, pb_task


@contextmanager
def zipdir(directory: Path) -> Path:
    _zip_dest = tempfile.mkdtemp()
    _zip_name = shutil.make_archive(
        base_name=directory.name,
        format='zip',
        root_dir=str(directory),
    )
    shutil.move(_zip_name, _zip_dest)
    yield Path(os.path.join(_zip_dest, os.path.basename(_zip_name)))
    shutil.rmtree(_zip_dest)



import asyncio
from functools import wraps

from .flow import CloudFlow


def asyncify(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        return asyncio.run(f(*args, **kwargs))

    return wrapper


@asyncify
async def deploy(args):
    await CloudFlow(
        path=args.path, name=args.name, workspace_id=args.workspace
    ).__aenter__()


@asyncify
async def status(args):
    from rich import box
    from rich.console import Console
    from rich.syntax import Syntax
    from rich.table import Table

    _t = Table('Attribute', 'Value', show_header=False, box=box.ROUNDED, highlight=True)

    console = Console()
    with console.status(f'[bold]Fetching status of {args.flow}...'):
        _result = await CloudFlow(flow_id=args.flow).status
        for k, v in _result.items():
            if k == 'yaml':
                v = Syntax(v, 'yaml', theme='monokai', line_numbers=True, code_width=40)
            else:
                v = str(v)
            _t.add_row(k, v)
        console.print(_t)


@asyncify
async def list(args):
    from rich import box
    from rich.console import Console
    from rich.table import Table

    _t = Table('ID', 'Status', 'Gateway', box=box.ROUNDED, highlight=True)

    console = Console()
    with console.status(f'[bold]Listing all flows...'):
        _result = await CloudFlow().list_all()
        if _result:
            for k in _result:
                _t.add_row(k['id'].split('-')[-1], k['status'], k['gateway'])
            console.print(_t)


@asyncify
async def remove(args):
    await CloudFlow(flow_id=args.flow).__aexit__()


@asyncify
async def logs(args):
    from rich.console import Console
    from .flow import pbar, pb_task

    with pbar:
        pbar.start_task(pb_task)
        pbar.update(
            pb_task,
            total=1,
            description=f'Press Ctrl-C to quit',
            title=f'Live streaming from {args.flow}',
        )
        await CloudFlow.logstream(params={'flow_id': args.flow})


@asyncify
async def login(args):
    from .auth import Auth

    await Auth.login()


def new(args):
    import os
    import shutil
    import sys

    rp = os.path.join(os.path.dirname(sys.modules['jcloud'].__file__), 'resources')

    shutil.copytree(os.path.join(rp, 'project-template'), os.path.abspath(args.path))

    from rich import print

    print(
        f'[green]New project create under {args.path}[/green]. Try [b]jc deploy {args.path}[/b].'
    )


def survey(args):
    # ask feedback
    from .auth import Survey

    Survey().ask(-1)



def main():
    from .parsers import get_main_parser
    import os

    args = get_main_parser().parse_args()

    if args.loglevel:
        os.environ['JCLOUD_LOGLEVEL'] = args.loglevel

    try:
        import threading
        from .helper import is_latest_version

        threading.Thread(
            target=is_latest_version, daemon=True, args=('jcloud',)
        ).start()
        from jcloud import api

        getattr(api, args.cli.replace('-', '_'))(args)
    except KeyboardInterrupt:
        pass


if __name__ == '__main__':
    main()



import json
import os
import webbrowser
from functools import lru_cache
from http.server import BaseHTTPRequestHandler, HTTPServer
from pathlib import Path
from typing import Dict
from urllib.parse import parse_qs
from urllib.request import Request, urlopen

import aiohttp

from .helper import get_logger
from rich import print

logger = get_logger()

JINA_CLOUD_CONFIG = 'config.json'


def _get_hub_config() -> Dict:
    hub_root = Path(os.environ.get('JINA_HUB_ROOT', Path.home().joinpath('.jina')))

    if not hub_root.exists():
        hub_root.mkdir(parents=True, exist_ok=True)

    config_file = hub_root.joinpath(JINA_CLOUD_CONFIG)
    if config_file.exists():
        with open(config_file) as f:
            return json.load(f)

    return {}


def _save_hub_config(config: Dict):
    hub_root = Path(os.environ.get('JINA_HUB_ROOT', Path.home().joinpath('.jina')))

    if not hub_root.exists():
        hub_root.mkdir(parents=True, exist_ok=True)

    config_file = hub_root.joinpath(JINA_CLOUD_CONFIG)
    with open(config_file, 'w') as f:
        json.dump(config, f)


@lru_cache()
def _get_cloud_api_url() -> str:
    """Get Cloud Api for transmiting data to the cloud.

    :raises RuntimeError: Encounter error when fetching the cloud Api Url.
    :return: Cloud Api Url
    """
    if 'JINA_HUBBLE_REGISTRY' in os.environ:
        return os.environ['JINA_HUBBLE_REGISTRY']
    else:
        try:
            req = Request(
                'https://api.jina.ai/hub/hubble.json',
                headers={'User-Agent': 'Mozilla/5.0'},
            )
            with urlopen(req) as resp:
                return json.load(resp)['url']
        except Exception as ex:
            logger.error(f'can not get hubble URL due to {ex!r}')


class Survey:
    def count(self) -> 'Survey':
        config = _get_hub_config()
        config['success_deploys'] = self.num_successful_deploys + 1
        _save_hub_config(config)
        return self

    @property
    def num_successful_deploys(self) -> int:
        config = _get_hub_config()
        return config.get('success_deploys', 0)

    @property
    def is_asked(self) -> bool:
        config = _get_hub_config()
        return bool(config.get('is_survey_done', False))

    def ask(self, threshold: int = 3) -> 'Survey':
        config = _get_hub_config()
        if threshold < 0 or (
            self.num_successful_deploys >= threshold and not self.is_asked
        ):
            from rich.prompt import Confirm
            from rich.markdown import Markdown

            is_survey = Confirm.ask(
                '[cyan]:bow: Would you like to take a quick survey about your user experience?\n'
                'It will only take [b]5 minutes[/b] but can help us understanding your usage better.[/cyan]\n'
                '[dim]a Google Form will be opened in your browser when typing [b]y[/b][/dim]'
            )
            if is_survey:
                webbrowser.open('https://forms.gle/1yEwNfh7pzeibxQy6', new=2)
                print(
                    Markdown(
                        '''
- If your browser does nothing, please [open this URL](https://forms.gle/1yEwNfh7pzeibxQy6).
- If your want to modify a submitted survey, or fill in later, please do `jc survey` in the terminal.
                '''
                    )
                )
                config['is_survey_done'] = True
                _save_hub_config(config)
            else:
                print(
                    '😅 No worries, when you are in better mood please do [b]jc survey[/b] in the terminal.'
                )
        return self


class Auth:
    @staticmethod
    def get_auth_token():
        config = _get_hub_config()
        return config.get('auth_token')

    @staticmethod
    async def login():
        api_host = _get_cloud_api_url()

        async with aiohttp.ClientSession() as session:
            redirect_url = 'http://localhost:8085'

            # TODO: add "app" parameter to API call
            async with session.get(
                url=f'{api_host}/v2/rpc/user.identity.authorize?provider=jina-login&prompt=login&redirectUri={redirect_url}'
            ) as response:
                response.raise_for_status()
                json_response = await response.json()
                logger.debug(json_response)
                webbrowser.open(json_response['data']['redirectTo'], new=2)

        done = False
        post_data = None

        class S(BaseHTTPRequestHandler):
            def _set_response(self):
                self.send_response(200)
                self.send_header('Content-type', 'text/html')
                self.end_headers()

            def do_POST(self):
                nonlocal done, post_data

                content_length = int(self.headers['Content-Length'])
                post_data = parse_qs(self.rfile.read(content_length))

                self._set_response()
                self.wfile.write(
                    'You have successfully logged in! You can close this window now.'.encode(
                        'utf-8'
                    )
                )
                done = True

            def log_message(self, format, *args):
                return

        server_address = ('', 8085)
        with HTTPServer(server_address, S) as httpd:
            logger.debug('Starting httpd...')
            while not done:
                httpd.handle_request()

        logger.debug(post_data)
        logger.debug('Stopping httpd...')

        async with aiohttp.ClientSession() as session:
            async with session.post(
                url=f'{api_host}/v2/rpc/user.identity.grant.auth0Unified',
                data=post_data,
            ) as response:
                response.raise_for_status()
                json_response = await response.json()
                logger.debug(json_response)
                # TODO: token will expire in 7 days but we need more time.
                token = json_response['data']['token']

        config = _get_hub_config()
        config['auth_token'] = token
        _save_hub_config(config)
        logger.debug('DONE')
        print('[green]:lock: You have successfully logged in![/green]')
